/* 
 * XJack - BeOS version of the xscreensaver XJack
 * BeOS rewrite, copyright © 1999 by John Yanarella (yanarejm@muohio.edu)
 *
 * version 1.1 (March 29th, 2000)
 *    - updated to compile under R5
 *
 * version 1.0 (November 12th, 1999)
 *    - original release
 *
 * TODO:
 *  - add ability to choose typeface from available fixed fonts 
 *  - add control for delay (inital value for m_delay)
 *  - control over colors (?)
 *
 * Although the BeOS "port" involved a fair amount of rewriting,
 * it's still covered by this license:
 * 
 * xscreensaver, Copyright (c) 1997 Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 *
 */

#include <TextView.h>
#include <stdlib.h>
#include <ctype.h>
#include "XJack.h"

static const char *source = "All work and no play makes Jack a dull boy.  ";

static const char* TITLE =
	"XJack for BeOS v1.1\n";
	
static const char* CREDITS = 
	"All work and no play makes Jack a dull boy.\n"
	"\nBeOS port by John Yanarella\n(yanarejm@muohio.edu)\n"
	"\nOriginal version by Jamie Zawinski\n(jwz@jwz.org)\n\n";

extern "C" _EXPORT BScreenSaver *instantiate_screen_saver(BMessage *message, image_id image)
{
	return new XJack(message, image);
}

XJack::XJack(BMessage *msg, image_id image)
 : BScreenSaver(msg, image)
{
}

XJack::~XJack() {
}

void XJack::StartConfig(BView *view)
{
	BTextView* creditsView = new BTextView(view->Bounds(), "credits", view->Bounds().InsetBySelf(5.0,0.0), B_FOLLOW_NONE, B_WILL_DRAW);
	
	creditsView->MakeEditable(false);
	creditsView->SetStylable(true);
	creditsView->MakeSelectable(false);	
	creditsView->SetAlignment(B_ALIGN_LEFT);
	creditsView->SetViewColor(216,216,216,0);
	
	BFont font(be_bold_font);
	font.SetSize(12.0);
	creditsView->SetFontAndColor(&font);
	creditsView->Insert(TITLE,strlen(TITLE));
	
	BFont font2(be_plain_font);
	font2.SetSize(10.0);
	creditsView->SetFontAndColor(&font2);	
	creditsView->Insert(CREDITS, strlen(CREDITS));
		
	view->AddChild(creditsView);
}


status_t XJack::StartSaver(BView * view, bool preview)
{
	SetTickSize(50000);
	
	view->SetLowColor(255,240,180);
	view->SetHighColor(0,0,0);
	view->FillRect(view->Bounds(), B_SOLID_LOW);

	// new seed every time we start
	time_t now;	
	now = time(NULL);
	srand(now);
	
	// Create a child view for drawing;
	// this is necessary because BView::CopyBits() (used to scroll data up)
	// implicitly calls BView::Draw() for the remaining region, so if we use it
	// in the Draw method of the default view we get "infinite" recursive 
	// calls to to BView::Draw().  Using a child BView, this works fine -- in fact
	// it erases the old region for us!!!
	m_myView = new BView(view->Bounds(), "XJack", B_FOLLOW_NONE, B_WILL_DRAW);
	
	m_myView->SetLowColor(255,240,180);
	m_myView->SetHighColor(0,0,0);
	m_myView->SetViewColor(255,240,180);
	
	BFont font(be_fixed_font);
	if (preview)
		font.SetSize(5.0);
	else
		font.SetSize(20.0);
	
	m_myView->SetFont(&font);

	view->AddChild(m_myView);
	
	m_mode = 0;
	m_hspace = m_vspace = 15;
	m_break_para = true;
	m_caps = false;
	m_sentences = m_paras = 0;	
	m_delay = 35000;

	font_height fheight;
	font.GetHeight(&fheight);
		
	m_char_width = (int)font.StringWidth("A");
	m_line_height = (int)(fheight.ascent + fheight.descent + 1);

	m_columns = (int)((m_myView->Bounds().Width() - m_hspace - m_hspace) / m_char_width);
	m_rows = (int)((m_myView->Bounds().Height() - m_vspace - m_vspace) / m_line_height);
	
	m_left = 0xFF & (rand() % ((m_columns / 2) +1));
	m_right = m_left + (0xFF & (rand() % (m_columns - m_left - 10) + 10));
	
	m_x = m_y = 0;
	
	m_s = source;
	
	return B_OK;
}

void XJack::Draw(BView *view, int32 frame)
{
	int word_length = 0;
	const char *s2;

	int delay = 0;
	
	for (s2 = m_s; *s2 && *s2 != ' '; s2++)
		word_length++;
	
	if (m_break_para || (*m_s != ' ' && (m_x + word_length) >= m_right))
	{
		m_x = m_left;
		m_y++;
		
		if (m_break_para)
			m_y++;
		
		if (m_mode == 1 || m_mode == 2)
		{
			// 1: left margin goes southwest, 2: southeast
			m_left += (m_mode == 1 ? 1 : -1);
			if (m_left >= m_right - 10)
			{
				if ((m_right < (m_columns - 10)) && (rand() & 1))
					m_right += (0xFF & (rand() % (m_columns - m_right)));
				else
					m_mode = 2;
			}
			else if (m_left <= 0)
			{
				m_left = 0;
				m_mode = 1;
			}
		}
		
		else if (m_mode == 3 || m_mode == 4)
		{
			// 3: right margin goes southeast, 4: southeast
			m_right += (m_mode == 3 ? 1 : -1);
			if (m_right >= m_columns)
			{
				m_right = m_columns;
				m_mode = 4;
			}
			else if (m_right <= m_left + 10)
			m_mode = 3;
		}
		
		if (m_y >= m_rows) // bottom of page
		{
			// scroll by 1-5 lines
			int lines = (rand() % 5 ? 0 : (0xFF & (rand() % 5))) + 1;
			if (m_break_para)
				lines++;
			
			// but sometimes scroll by a whole page
			if ((rand() % 100) == 0)
				lines += m_rows;
			
			while (lines > 0)
			{
				int inc = m_line_height / 7;
				
				if (inc <= 0) inc = 1;
				
				for (int i = 0; i < m_line_height; i += inc) 
				{
					if (i > m_line_height)
						i = m_line_height;
					
					BRect bounds = m_myView->Bounds();
					BRect source(0, inc, bounds.Width(), bounds.Height() - inc);
					BRect dest(source);
					dest.OffsetTo(0,0);
					m_myView->CopyBits(source,dest);
					m_myView->Sync();
				}
				
				m_y--;
				lines--;
				
				m_columns = (int)((m_myView->Bounds().Width() - m_hspace - m_hspace) / m_char_width);
				m_rows = (int)((m_myView->Bounds().Height() - m_vspace - m_vspace) / m_line_height);
				
				if (m_y > m_rows) 
					m_y = m_rows - 1;
				if (m_x > m_columns) 
					m_x = m_columns-2;
				
				if (m_delay)
					delay += m_delay;
			}
			
			if (m_y < 0)
				m_y = 0;
		}
		
		m_break_para = false;		
	}

	if (*m_s != ' ')
	{
		char c = *m_s;
		int xshift = 0, yshift = 0;
		if ((rand() % 50) == 0)
		{
			xshift = rand() % ((m_char_width / 3) + 1); // mis-strike
			yshift = rand() % ((m_line_height / 6) + 1);
			if ((rand() % 3) == 0)
				yshift *= 2;
			if (rand() & 1)
				xshift = -xshift;
			if (rand() & 1)
				yshift = -yshift;
		}
		
		if ((rand() % 250) == 0) // adjascent-key typo
		{
			static const char * const typo[] = {
				"asqw", "ASQW", "bgvhn", "cxdfv", "dserfcx", "ewsdrf",
				"Jhuikmn", "kjiol,m", "lkop;.,", "mnjk,", "nbhjm", "oiklp09",
				"pol;(-0", "redft54", "sawedxz", "uyhji87", "wqase32",
				"yuhgt67", ".,l;/", 0 };
			int i = 0;
			
			while (typo[i] && typo[i][0] != c)
				i++;
			if (typo[i])
				c = typo[i][0xFF & (rand() % strlen(typo[i]+1))];
	    }

		// caps typo
	  	if (c >= 'a' && c <= 'z' && (m_caps || (rand() % 350) == 0))
		{
			c -= ('a'-'A');
			if (c == 'O' && rand() & 1)
				c = '0';
	    }

	OVERSTRIKE:
		BPoint p((m_x * m_char_width) + m_hspace + xshift, 
		         (m_y * m_line_height) + m_vspace + m_vspace + yshift);
	  	m_myView->DrawChar(c, p);
	  	m_myView->Sync();
		       		
		if (xshift == 0 && yshift == 0 && (0 == (rand() & 3000)))
		{
			if (rand() & 1)
				xshift--;
			else
				yshift--;
			goto OVERSTRIKE;
		}

		if ((tolower(c) != tolower(*m_s))
			? ((rand() % 10) == 0)		// backup to correct 
			: ((rand() % 400) == 0))	// fail to advance 
		{
			m_x--;
			m_s--;
		
			if (m_delay) 
				delay += 0xFFFF & (m_delay + (rand() % (m_delay * 10)));
	    }
	}

	m_x++;
    m_s++;

	if (rand() % 200 == 0)
	{
		if (rand() & 1 && m_s != source)
	    	m_s--;	// duplicate character
	  	else if (*m_s)
	  		m_s++;	// skip character 
	}

	if (*m_s == 0)
	{
		m_sentences++;
		m_caps = (rand() % 40 == 0);	// capitalize sentence 

		if ((rand() % 10) == 0 ||	// randomly break paragraph 
			(m_mode == 0 &&
			(((rand() % 10) == 0) || m_sentences > 20)))
		{
			m_break_para = true;
			m_sentences = 0;
			m_paras++;

			if (rand() & 1)		// mode = 0 50% of the time 
				m_mode = 0;
			else
				m_mode = (0xFF & (rand() % 5));

			if ((rand() % 2) == 0)	// re-pick margins 
			{
				m_left = 0xFF & (rand() % (m_columns / 3));
				m_right = m_columns - (0xFF & (rand() % (m_columns / 3)));

				if (rand() % 3 == 0)	// sometimes be wide 
					m_right = m_left + ((m_right - m_left) / 2);
			}

			if (m_right - m_left <= 10)	// introduce sanity
			{
				m_left = 0;
				m_right = m_columns;
			}

			if (m_right - m_left > 50)	// if wide, shrink and move 
			{
		  		m_left += (0xFF & (rand() % ((m_columns - 50) + 1)));
				m_right = m_left + (0xFF & ((rand() % 40) + 10));
			}

			if (m_mode == 0 &&
				m_right - m_left < 25 &&
				m_columns > 40)
			{
				m_right += 20;
				if (m_right > m_columns)
					m_left -= (m_right - m_columns);
			}
		}
		m_s = source;
	}

	if (m_delay)
	{
		delay += m_delay;
		if (rand() % 3 == 0)
			delay += 0xFFFFFF & ((rand() % (m_delay * 5)) + 1);

		if (m_break_para)
			delay += 0xFFFFFF & ((rand() % (m_delay * 15)) + 1);
	}

	if (m_delay && delay > 0)
		SetTickSize(delay);
}
